/*============================================================================
 * Interpolation by cubic splines in n dimensions.
 * Works with data located w.r.t. regular grids possibly stretched.
 * It accepts extrapolation but the user should be aware that extrapolating
 * from 3rd order polynomials can quickly lead to absurd values. 
 * The functions cub_interpolate_1 & cub_interpolate_n are from: 
 * http://www.paulinternet.nl/?page=bicubic (november 2015)
 * It is also possible to select a simpler linear interpolation.
 *============================================================================*/

/* Code_Saturne version 4.2-alpha */

/*
  This file is part of Code_Saturne, a general-purpose CFD tool.

  Copyright (C) 1998-2015 EDF S.A.

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
  Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * Standard C library headers
 *----------------------------------------------------------------------------*/

#include <math.h>

/*----------------------------------------------------------------------------
 * Local headers
 *----------------------------------------------------------------------------*/

#include "bft_mem.h"

#include "splines.h"

#define min(x,y) ((x) < (y) ? (x) : (y))

/*----------------------------------------------------------------------------*/
/*!
 * \brief  Linear interpolation 1-D
 */
/*----------------------------------------------------------------------------*/

double lin_interpolate_1 (double pi[2], 
			  double xi) 
{
  return xi*pi[1] + (1.-xi)*pi[0];
}

/*----------------------------------------------------------------------------*/
/*!
 * \brief  Linear interpolation N-D
 */
/*----------------------------------------------------------------------------*/

double lin_interpolate_n (int     ndim, 
			  double *pi,
			  double *xi)
{
  if (ndim == 1) 
    return lin_interpolate_1(pi, *xi);

  else {
    double arr[2];
    int ic = pow(2,ndim-1);

    arr[0] = lin_interpolate_n(ndim-1, pi,      xi+1);
    arr[1] = lin_interpolate_n(ndim-1, pi+ic,   xi+1);
    return lin_interpolate_1(arr, *xi);
  }
}

/*----------------------------------------------------------------------------*/
/*!
 * \brief  Cubic interpolation 1-D
 */
/*----------------------------------------------------------------------------*/

double cub_interpolate_1 (double pi[4], 
			  double xi)
{
  return pi[1] + 0.5*xi*(pi[2]-pi[0] + xi*(2.*pi[0]-5.*pi[1]+4.*pi[2]-pi[3] + xi*(3.*(pi[1]-pi[2])+pi[3]-pi[0])));
}

/*----------------------------------------------------------------------------*/
/*!
 * \brief  Cubic interpolation N-D
 */
/*----------------------------------------------------------------------------*/

double cub_interpolate_n (int     ndim, 
			  double *pi, 
			  double *xi)
{
  if (ndim == 1) 
    return cub_interpolate_1(pi, *xi);

  else {
    double arr[4];
    int ic = pow(4,ndim-1);

    arr[0] = cub_interpolate_n(ndim-1, pi,      xi+1);
    arr[1] = cub_interpolate_n(ndim-1, pi+ic,   xi+1);
    arr[2] = cub_interpolate_n(ndim-1, pi+2*ic, xi+1);
    arr[3] = cub_interpolate_n(ndim-1, pi+3*ic, xi+1);
    return cub_interpolate_1(arr, *xi);
  }
}

/*----------------------------------------------------------------------------*/
/*!
 * \brief  Find the rank in each dimension from the global index
 */
/*----------------------------------------------------------------------------*/

int* find_indices(int  ndim, 
		  int  ix, 
		  int *nx)
{
  int *rx; //rx[i] contains the rank in i-th dir.
  int i1, i2, is, id;

  BFT_MALLOC(rx  , ndim,        int);

  is = 1;
  for (id = 1; id < ndim; id++)
    is *= nx[id];
  rx[0] = ix / is;

  if (ndim > 2) {
    for (id = 1; id < ndim-1; id++) {
      i1 = ndim - id;
      is = 1;
      for (i2 = 1; i2 < i1; i2++)
	is *= nx[i1 - i2];
      rx[id] = ix / is % nx[id];
    }
  }
  if (ndim > 1)
    rx[ndim-1] = ix % nx[ndim-1];

  return rx;
  BFT_FREE(rx);
}

/*----------------------------------------------------------------------------*/
/*!
 * \brief  Prepare data for interpolation 
 *
 * \param [in]  cubint : interpolation with cubic splines if set to 1, linear otherwise
 * \param [in]  dernull : first derivatives null at boudaries if set to 1, second derivatives
 *     null otherwise
 * \param [in]  stretched : if the grid made by the locations of the data to be interpolated is
 *     stretched, this parameter should be set to 1, otherwise set to 0, a step can 
 *     be avoided in this function
 * \param [in]  nd : number of physical dimensions of the problem (between 1 and infty)
 * \param [in]  ng : number of points in the data to be interpolated IN EACH DIMENSION
 * \param [in]  xglob : locations of the points aforementioned. This has to be carefully organized.
 *     Considering ntot the total number of points 
 *     - in 2d (xi,yi)    -> (x(i+ntot), x(i))
 *     - in 3d (xi,yi,zi) -> (x(i+2*ntot), x(i+ntot), x(i)).
 *     Also the locations are assumed to be sorted from min to max or max to min.
 * \param [in]  pglob : values at the points aforementioned
 * \param [in]  ni : number of points where interpolated values have to be calculated
 * \param [in]  xintp : locations of the points aforementioned
 * \param [out] pintp : values at the points aforementioned
 */
/*----------------------------------------------------------------------------*/

void cs_cub_interpolate(int     cubint, 
			int     dernull, 
			int     stretched, 
			int     ndim,
			int    *ng, 
			double *xglob, 
			double *pglob, 
			int     sizepi, 
			double *xintp, 
			double *pintp)
{ 
  int id, ig, is, il, ia, ir, ic, ii, i1;
  int mini;
  int sizepg, sizepa, sizepl; // size of pglob, paugm, plocal 
  double l1, l2, d1; // lengths used in corrections on stretched grids
  int *na; //na[i] contains the number of points in i-th dir. , augmented
  int *rx; //rx[i] contains the rank in i-th dir.
  int *ra; //ra[i] contains the rank in i-th dir. , augmented
  int *rg; //rg[i] contains the values to switch from rank number to total numbering
  int *rl; //rr[i] contains the values to switch from rank number to total numbering in 4**ndim neighborood
  int *iz; //iz[i] contains the directions in which augmentation is necessary
  int *np; //np simply contains ndim 4; 
  int *rp; //rp simply contains the increments to get the neighborood of interp point
  int *rs; //scratch
  double *paugm; // data with extrapolated values to have sufficient neighborood 
  double *xaugm; // locations of data points
  double *ploc; // local neighborood of location xintp 
  double *xloc; // local x vector
  double *xsch; // x vector for the search

  if (cubint)
    sizepl = pow(4,ndim);
  else 
    sizepl = pow(2,ndim); 

  BFT_MALLOC(na  , ndim,        int);
  BFT_MALLOC(rx  , ndim,        int);
  BFT_MALLOC(ra  , ndim,        int);
  BFT_MALLOC(rg  , ndim,        int);
  BFT_MALLOC(rl  , ndim,        int);
  BFT_MALLOC(rs  , ndim,        int);
  BFT_MALLOC(np  , ndim,        int);
  BFT_MALLOC(iz  , ndim,        int);
  BFT_MALLOC(rp  , ndim*sizepl, int);
  BFT_MALLOC(ploc, sizepl,      double);
  BFT_MALLOC(xloc, ndim*sizepl, double);
  BFT_MALLOC(xsch, ndim,        double);

  sizepg = 1;
  sizepa = 1;
  for (id = 0; id < ndim; id++) {
    na[id] = ng[id]+2;
    sizepg *= ng[id];
    sizepa *= na[id];
  }

  BFT_MALLOC(paugm, sizepa,      double);
  BFT_MALLOC(xaugm, ndim*sizepa, double);	     
 
  // Calculations of rg, ra and rl to be able to find neighboors from the global index
  for (id = 0; id < ndim; id++) {
    rg[id] = 1;
    ra[id] = 1;
    rl[id] = 1;
    for (i1 = ndim-1; i1 > id; i1--) {
      rg[id] *= ng[i1];
      ra[id] *= na[i1];
      if (cubint)
	rl[id] *= 4;
      else
	rl[id] *= 2;	
    }
  }

  if (cubint) {
    for (ia = 0; ia < sizepa; ia++) {

      rx = find_indices(ndim, ia, na);

      ig = 0;
      for (id = 0; id < ndim; id++) {
	rx[id] -= 1;
	ig += rg[id]*rx[id];
      }

      // Calculation of the "augmented grid" to deal with boundary conditions
      ir = 0;
      for (id = 0; id < ndim; id++) 
	if (rx[id] == -1 || rx[id] == ng[id]) 
	  ir += 1;
    
      if (ir == 0) {
	paugm[ia] = pglob[ig];
	for (id = 0; id < ndim; id++)
	  xaugm[id*sizepa+ia] = xglob[id*sizepg+ig];
      }

      if (ir > 0) {
	i1 = 0;
	for (id = 0; id < ndim; id++) {
	  if (rx[id] == -1 || rx[id] == ng[id]) {
	    iz[i1] = id;
	    i1++;
	  }
	}
	ic = 0;
	for (i1 = 0; i1 < ir; i1++) {
	  if (rx[iz[i1]] == -1)
	    ic += rg[iz[i1]];
	  else if (rx[iz[i1]] == ng[iz[i1]])
	    ic -= rg[iz[i1]];	  
	}
	for (id = 0; id < ndim; id++)
	  xaugm[id*sizepa+ia] = 2*xglob[id*sizepg+ig+ic]-xglob[id*sizepg+ig+2*ic];
	if (dernull == 1)
	  paugm[ia] = pglob[ig+2*ic];
	else {
	  mini = 3;
	  for (id = 0; id < ndim; id++)
	    mini = min(mini, ng[id]);
	  if  (mini < 3)
	    paugm[ia] = 2*pglob[ig+ic] - pglob[ig+2*ic];
	  else {
	    l1 = 0.;
	    l2 = 0.;
	    for (id = 0; id < ndim; id++) {
	      l1 += pow(xglob[id*sizepg+ig+2*ic] - xglob[id*sizepg+ig+ic], 2);
	      l2 += pow(xglob[id*sizepg+ig+3*ic] - xglob[id*sizepg+ig+ic], 2);
	    }
	    l1 = sqrt(l1);
	    l2 = sqrt(l2);
	    d1 = (pglob[ig+3*ic] - pglob[ig+ic]) * l1 / l2;
	    paugm[ia] = d1 + 3*pglob[ig+ic] - 2*pglob[ig+2*ic];
	  }
	}
      }
    }
  }

  // Calculations of rp to be able to find neighboors with the local index
  for (il = 0; il < sizepl; il++) {
    for (id = 0; id < ndim; id++) {
      if (cubint) 
	np[id] = 4;
      else
	np[id] = 2;
    }
    rs = find_indices(ndim, il, np);
    for (id = 0; id < ndim; id++)
      rp[il*ndim+id] = rs[id];
  }

  // Calculation of the local neighborood of size 4**ndim
  for (ii = 0; ii < sizepi; ii++) {
    // Calculation of the search location in the local neighborood
    for (id = 0; id < ndim; id++) {
      is = 0;
      ig = 0;
      do {
	xsch[id] = (xintp[id*sizepi+ii]-xglob[id*sizepg+ig]) 
	  / (xglob[id*sizepg+ig+rg[id]]-xglob[id*sizepg+ig]);
	is += 1; 
	ig = is*rg[id];  
      } while (xintp[id*sizepi+ii] >= xglob[id*sizepg+ig] && is < ng[id]-1);
      rx[id] = is-1;
    }

    // By default the local neighborood is directly obtained from the augmented data
    if (cubint) {
      for (il = 0; il < sizepl; il++) {
	ia = 0;
	for (id = 0; id < ndim; id++)
	  ia += ra[id]*(rx[id]+rp[il*ndim+id]);
	ploc[il] = paugm[ia];
	for (id = 0; id < ndim; id++)
	  xloc[id*sizepl+il] = xaugm[id*sizepa+ia];
      }
    }
    else {
      for (il = 0; il < sizepl; il++) {
	ig = 0;
	for (id = 0; id < ndim; id++)
	  ig += rg[id]*(rx[id]+rp[il*ndim+id]);
	ploc[il] = pglob[ig];
	for (id = 0; id < ndim; id++)
	  xloc[id*sizepl+il] = xglob[id*sizepa+ig];
      }
    }

    if (stretched && cubint) {       
      // if the grid is stretched, it is necessary to use modified values for the data in the
      // local neighborood since further the grid is supposed to be regular.

      for (il = 0; il < sizepl; il++) {
	rs = find_indices(ndim, il, np);
	ir = 0;
	for (id = 0; id < ndim; id++)
	  if (rs[id] == 0 || rs[id] == 3)
	    ir += 1;
 
	if (ir > 0) { // if we are at a boundary
	  i1 = 0;
	  for (id = 0; id < ndim; id++) {
	    if (rs[id] == 0 || rs[id] == 3) {
	      iz[i1] = id;
	      i1++;
	    }
	  }
	  ic = 0;
	  for (i1 = 0; i1 < ir; i1++) {
	    if (rs[iz[i1]] == 0)
	      ic += rl[iz[i1]];
	    else if (rs[iz[i1]] == 3)
	      ic -= rl[iz[i1]];	  
	  }

	  l1 = 0.;
	  l2 = 0.;
	  for (id = 0; id < ndim; id++) {
	    l1 += pow((xloc[id*sizepl+il]    - xloc[id*sizepl+il+ic]),   2);
	    l2 += pow((xloc[id*sizepl+il+ic] - xloc[id*sizepl+il+2*ic]), 2);
	  }
	  l1 = sqrt(l1);
	  l2 = sqrt(l2);
	  ploc[il] = (ploc[il] - ploc[il+ic]) * l2 / l1 + ploc[il+ic]; // (p4-p3) * (x3-x2)/(x4-x3) + p3
	}
      }
    }
    // Calculation of the inporlate from the local neighborood and the search location in this neighborood
    if (cubint)
      pintp[ii] = cub_interpolate_n(ndim, ploc, xsch);
    else
      pintp[ii] = lin_interpolate_n(ndim, ploc, xsch);
  }

  BFT_FREE(na);
  BFT_FREE(rx);
  BFT_FREE(ra);
  BFT_FREE(rg);
  BFT_FREE(rl);
  BFT_FREE(rs);
  BFT_FREE(np);
  BFT_FREE(iz);
  BFT_FREE(rp);
  BFT_FREE(ploc);
  BFT_FREE(xloc);
  BFT_FREE(xsch);
  BFT_FREE(paugm);
  BFT_FREE(xaugm);
}
